package main;

import meerkat.crypto.mixnet.Mix2ZeroKnowledgeVerifier;
import meerkat.crypto.mixnet.Mixer;
import meerkat.protobuf.Crypto;
import meerkat.protobuf.Mixing;
import qilin.util.Pair;

import java.util.Arrays;
import java.util.List;


/**
 * Created by Tzlil on 12/17/2015.
 */
public class MainMixing {

    Mixer mixer;
    Mix2ZeroKnowledgeVerifier verifier;
    int n,layers;

    public MainMixing(Mixer mixer, Mix2ZeroKnowledgeVerifier verifier, int n) {
        this.mixer = mixer;
        this.verifier = verifier;
        this.n = n;
        this.layers = (int) (2 * Math.log(n) / Math.log(2)) - 1;  // layers = 2logn -1
    }

    private void firstMixing( ) throws Exception
    {
        updateBB(mixer.mix(getInputForFirstMixer()));
    }

    private void otherMixing() throws Exception
    {
        List<Mixing.ZeroKnowledgeProof[][]> zeroKnowledgeProofsLists  = downloadProofs();
        List<Crypto.RerandomizableEncryptedMessage[][]> rerandomizableEncryptedMessageLists  = downloadEncryptionTable();

        for (int i = 0; i < zeroKnowledgeProofsLists.size(); i++)
        {
            if(!verifyTable(n,layers,zeroKnowledgeProofsLists.get(i),rerandomizableEncryptedMessageLists.get(i)))
                throw new Exception();
        }

        Crypto.RerandomizableEncryptedMessage[] lastLayerInCurrent,firstLineInNext;
        for (int i = 0; i < rerandomizableEncryptedMessageLists.size() - 1 ; i++)
        {
            lastLayerInCurrent = rerandomizableEncryptedMessageLists.get(i)[layers - 1];
            firstLineInNext = rerandomizableEncryptedMessageLists.get(i + 1)[0];
            if(!Arrays.equals(lastLayerInCurrent,firstLineInNext))
                throw new Exception();
        }

        List<Crypto.RerandomizableEncryptedMessage> inputForMixer =
                Arrays.asList(rerandomizableEncryptedMessageLists
                        .get(rerandomizableEncryptedMessageLists.size() - 1)[layers - 1]);

        updateBB(mixer.mix(inputForMixer));
    }

    private List<Mixing.ZeroKnowledgeProof[][]> downloadProofs()
    {
        return null;
    }

    private List<Crypto.RerandomizableEncryptedMessage[][]> downloadEncryptionTable()
    {
        return null;
    }

    private List<Crypto.RerandomizableEncryptedMessage> getInputForFirstMixer()
    {
        return null;
    }

    private void updateBB(Pair<Mixing.ZeroKnowledgeProof[][],Crypto.RerandomizableEncryptedMessage[][]> mixerOutput){

    }

    private boolean verifyTable(int n,int layers, Mixing.ZeroKnowledgeProof[][] zeroKnowledgeProofs,
                                Crypto.RerandomizableEncryptedMessage[][] rerandomizableEncryptedMessages)
    {
        int index1,index2,layer;

        //initialize locationChecksum table
        // use for check BeneshNet validity
        boolean[][] locationChecksum = new boolean[layers][n];
        for (boolean[] locationChecksumLayer: locationChecksum) {
            Arrays.fill(locationChecksumLayer,false);
        }


        for (Mixing.ZeroKnowledgeProof[] zkpLayer: zeroKnowledgeProofs) {
            for (Mixing.ZeroKnowledgeProof zkp: zkpLayer) {
                Mixing.ZeroKnowledgeProof.Location location = zkp.getLocation();
                index1 = location.getI();
                index2 = location.getJ();
                layer = location.getLayer();

                // check location validity
                if (layer > layers >> 1) {
                    if (index2 - index1 != n >> (layers - layer))
                        return false;
                }
                else{
                    if (index2 - index1 != n >> (layer + 1))
                        return false;
                }
                
                // mark location in table
                locationChecksum[layer][index1] = true;
                locationChecksum[layer][index2] = true;
                
                // verify proof
                if(!verifier.verify(rerandomizableEncryptedMessages[index1][layer],
                        rerandomizableEncryptedMessages[index2][layer],
                        rerandomizableEncryptedMessages[index1][layer + 1],
                        rerandomizableEncryptedMessages[index2][layer + 1],
                        zkp))
                    return  false;
            }
        }

        // verify all necessary locations for BeneshNet were proved
        for (boolean[] checksumLayer: locationChecksum) {
            for (boolean locationBoolean: checksumLayer) {
                if (!locationBoolean)
                    return false;
            }
        }
        return true;
    }
}
