package mixer;

import java.util.ArrayList;
import java.util.List;
import qilin.util.Pair;
import java.util.Random;

import com.google.protobuf.InvalidProtocolBufferException;

import meerkat.protobuf.Crypto.*;
import meerkat.protobuf.Mixing.*;

import meerkat.crypto.Encryption;
import meerkat.crypto.mixnet.Mix2ZeroKnowledgeProver;

public class Mixer implements meerkat.crypto.mixnet.Mixer {

	private Random random;
	private Mix2ZeroKnowledgeProver prover;
	private Encryption encryptor;

	public Mixer(Random rand, Mix2ZeroKnowledgeProver prov, Encryption enc) {
		this.random = rand;
		this.prover = prov;
		this.encryptor = enc;

	}

	public Pair<ZeroKnowledgeProof[][], RerandomizableEncryptedMessage[][]> mix(List<RerandomizableEncryptedMessage> ciphertexts) throws InvalidProtocolBufferException {

		int n = ciphertexts.size();
		int nDiv2 = n >> 1;
		// assert n = 2^k and n > 1
		if (n <= 1 || ((n & (n - 1)) != 0))
			return null;

		//initialization
		int layers = (int) (2 * Math.log(n) / Math.log(2)) - 1;  // layers = 2logn -1
		RerandomizableEncryptedMessage[][] encryptionTable = new RerandomizableEncryptedMessage[layers + 1][n];
		ZeroKnowledgeProof[][] proofsTable = new ZeroKnowledgeProof[layers][nDiv2];
		int index1, index2, switchIndex = 0;
		EncryptionRandomness r1, r2;
		RerandomizableEncryptedMessage e1, e2;
		boolean half = true;
		MixNetwork mixNetwork = new MixNetwork(new RandomPermutation(n,random));
		Switch[] switchesLayer;

		// set first level of encryption
		for (int j = 0; j < n; j++) {
			encryptionTable[0][j] = ciphertexts.get(j);
		}

		// main loop
		for (int layer = 0; layer < layers; layer++)
		{
			switchesLayer = mixNetwork.getSwitchesByLayer(layer);
			for (Switch sw : switchesLayer) {
				index1 = sw.i;
				index2 = sw.j;
				e1 = encryptionTable[layer][index1];
				e2 = encryptionTable[layer][index2];
				r1 = encryptor.generateRandomness(random);
				r2 = encryptor.generateRandomness(random);
				if (!sw.value) {
					encryptionTable[layer + 1][index1] = encryptor.rerandomize(e1, r1);
					encryptionTable[layer + 1][index2] = encryptor.rerandomize(e2, r2);

				} else {
					encryptionTable[layer + 1][index1] = encryptor.rerandomize(e2, r2);
					encryptionTable[layer + 1][index2] = encryptor.rerandomize(e1, r1);
				}
				proofsTable[layer][switchIndex] =
						prover.prove(e1, e2, encryptionTable[layer + 1][index1],
								encryptionTable[layer + 1][index2],
								sw.value, sw.i,sw.j,sw.layer, r1, r2);
			}
		}
		return new Pair<ZeroKnowledgeProof[][], RerandomizableEncryptedMessage[][]>(proofsTable, encryptionTable);
	}

}