package verifier;

import meerkat.crypto.mixnet.Mix2ZeroKnowledgeVerifier;
import meerkat.protobuf.ConcreteCrypto.ElGamalCiphertext;
import meerkat.protobuf.Crypto;
import meerkat.protobuf.Mixing;
import necessary.General;
import qilin.primitives.Group;

import java.math.BigInteger;


public class Verifier implements Mix2ZeroKnowledgeVerifier {

    /**
     * Return true iff the proof is valid.
     * @param in1
     * @param in2
     * @param out1
     * @param out2
     * @return
     */

    Group group;
    General general;

    public Verifier(Group group,General general) {
        this.group = group;
        this.general = general;
    }

    public boolean verify(Crypto.RerandomizableEncryptedMessage in1,
                          Crypto.RerandomizableEncryptedMessage in2,
                          Crypto.RerandomizableEncryptedMessage out1,
                          Crypto.RerandomizableEncryptedMessage out2,
                          Mixing.ZeroKnowledgeProof proof)
    {

        ElGamalCiphertext e1ElGamal = general.calcRerandomizable2ElGamal(in1);
        ElGamalCiphertext e2ElGamal = general.calcRerandomizable2ElGamal(in2);
        ElGamalCiphertext e1TagElGamal = general.calcRerandomizable2ElGamal(out1);
        ElGamalCiphertext e2TagElGamal = general.calcRerandomizable2ElGamal(out2);


        return verifyElGamaOrProof(e1ElGamal,e2ElGamal,e1TagElGamal,e2TagElGamal,proof.getFirst()) &&
               verifyElGamaOrProof(e1ElGamal,e2ElGamal,e1TagElGamal,e2TagElGamal,proof.getSecond()) &&
               verifyElGamaOrProof(e1ElGamal,e2ElGamal,e1TagElGamal,e2TagElGamal,proof.getThird()) &&
               verifyElGamaOrProof(e1ElGamal,e2ElGamal,e1TagElGamal,e2TagElGamal,proof.getFourth());
    }

    public boolean verifyElGamaOrProof(ElGamalCiphertext e1,
                                 ElGamalCiphertext e2,
                                 ElGamalCiphertext e1New,
                                 ElGamalCiphertext e2New,
                                 Mixing.ZeroKnowledgeProof.OrProof orProof)
    {

        Mixing.ZeroKnowledgeProof.OrProof.ForRandomOracle forRandomOracle =
                Mixing.ZeroKnowledgeProof.OrProof.ForRandomOracle.newBuilder()
                        .setG1(orProof.getG1())
                        .setH1(orProof.getH1())
                        .setG2(orProof.getG2())
                        .setH2(orProof.getH2())
                        .setG1Tag(orProof.getG1Tag())
                        .setH1Tag(orProof.getH1Tag())
                        .setG2Tag(orProof.getG2Tag())
                        .setH2Tag(orProof.getH2Tag())
                        .setU(orProof.getU())
                        .setV(orProof.getV())
                        .setUTag(orProof.getUTag())
                        .setVTag(orProof.getVTag())
                        .build();

        return   //input
                 orProof.getG1().equals(general.getG())&&
                 orProof.getH1().equals(group.add(e1New.getC1(), group.negate(e1.getC1())))&&
                 orProof.getG2().equals(general.getH())&&
                 orProof.getH2().equals(group.add(e1New.getC2(), group.negate(e1.getC2())))&&
                 // input'
                 orProof.getG1Tag().equals(general.getG())&&
                 orProof.getH1Tag().equals(group.add(e2New.getC1(), group.negate(e2.getC1())))&&
                 orProof.getG2Tag().equals(general.getH())&&
                 orProof.getH2Tag().equals(group.add(e2New.getC2(), group.negate(e2.getC2()))) &&
                 // hash
                 // assert (c1 + c2 ) % group size == hash (imput + step1) % group size
                 new BigInteger(orProof.getC1().toByteArray()).add(new BigInteger(orProof.getC2().toByteArray())).mod(group.orderUpperBound())
                 .equals(general.hash(forRandomOracle).mod(group.orderUpperBound()).mod(group.orderUpperBound()))&&
                 // proof
                 // g1 ^ z == u * ( h1 ^ c1 ) && g2 ^ z == v * ( h2 ^ c1 ) && the same for tag case
                 group.multiply(orProof.getG1(), new BigInteger(orProof.getZ().toByteArray()))
                         .equals(group.add(orProof.getU(), group.multiply(orProof.getH1(),new BigInteger(orProof.getC1().toByteArray())))) &&
                 group.multiply(orProof.getG2(), new BigInteger(orProof.getZ().toByteArray()))
                         .equals(group.add(orProof.getV(), group.multiply(orProof.getH2(),new BigInteger(orProof.getC1().toByteArray())))) &&
                 group.multiply(orProof.getG1Tag(), new BigInteger(orProof.getZTag().toByteArray()))
                         .equals(group.add(orProof.getUTag(), group.multiply(orProof.getH1Tag(),new BigInteger(orProof.getC2().toByteArray())))) &&
                 group.multiply(orProof.getG2Tag(), new BigInteger(orProof.getZTag().toByteArray()))
                         .equals(group.add(orProof.getVTag(), group.multiply(orProof.getH2Tag(),new BigInteger(orProof.getC2().toByteArray()))));

    }
}
